#ifndef LFT_ICMPTRACE_H
#define LFT_ICMPTRACE_H

#include "lft_lib.h"

#define ICMP_TRACE	30

int generateICMPPacket(
	lft_session_params * sess,
	LFT_CALLBACK err,
	LFT_CALLBACK evt,
	struct icmp_trace_packet_s * packet,
	u_char ttlval,
	u_short unique_id,
	u_short seq);
void icmp_trace_main_loop(lft_session_params * sess, LFT_CALLBACK err, LFT_CALLBACK evt);
#if !defined(__CYGWIN__) && !defined(WIN32) && !defined(_WIN32)
void pcap_icmp_process_packet(u_char *user_data, const struct pcap_pkthdr *hdr, const u_char *packet);
unsigned int icmp_pmtud_send_verify_hop(lft_session_params *sess, short nhop);
#endif
#endif
